# Copyright 2011-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the web resources of the testkeyserver."""

import os
import shutil
import urllib

import doctest
import fixtures
from testtools import TestCase
from testtools.deferredruntest import AsynchronousDeferredRunTest
from testtools.matchers import (
    Contains,
    DocTestMatches,
)
from twisted.internet.endpoints import serverFromString
from twisted.python.failure import Failure
from twisted.web.client import getPage
from twisted.web.server import Site

from gpgservice_client.testing.keyserver.harness import KEYS_DIR
from gpgservice_client.testing.keyserver.web import KeyServerResource

__metaclass__ = type


class RegularCallbackExecuted(Exception):
    """Raised if a regular Twisted callback is called when the request
    is supposed to return an HTTP error."""


class TestWebResources(TestCase):

    run_tests_with = AsynchronousDeferredRunTest.make_factory(timeout=2)

    def setUpKeysDirectory(self):
        path = os.path.join(self.useFixture(fixtures.TempDir()).path, 'keys')
        shutil.copytree(KEYS_DIR, path)
        return path

    def makeService(self):
        """Run a test key server on whatever port we have available."""
        from twisted.internet import reactor
        resource = KeyServerResource(self.setUpKeysDirectory())
        site = Site(resource)
        endpoint = serverFromString(reactor, 'tcp:0')
        return endpoint.listen(site)

    def fetchResource(self, listening_port, path, method):
        """GET the content at 'path' from the web server at 'listening_port'.
        """
        url = 'http://localhost:%s/%s' % (
            listening_port.getHost().port,
            path.lstrip('/'))
        return getPage(url, method=method)

    def getURL(self, path):
        """Start a test key server and get the content at 'path'."""
        d = self.makeService()

        def service_started(port):
            self.addCleanup(port.stopListening)
            return self.fetchResource(port, path, 'GET')
        return d.addCallback(service_started)

    def assertContentMatches(self, path, content):
        """Assert that the key server content at 'path' matches 'content'."""
        d = self.getURL(path)
        flags = (
            doctest.REPORT_NDIFF | doctest.ELLIPSIS |
            doctest.NORMALIZE_WHITESPACE)
        return d.addCallback(self.assertThat, DocTestMatches(content, flags))

    def assertRaises404ErrorForKeyNotFound(self, path):
        """Assert that the test server returns a 404 response
        for attempts to retrieve an unknown key.
        ."""
        d = self.getURL(path)

        def regular_execution_callback(content):
            # A really Twisted(tm) error check:
            #
            # This callback should _not_ be called, because setting
            # the HTTP status code to 500 in Lookup.processRequest()
            # prevents this. On the other hand, if the status code is
            # _not_ set, the callback check_error_details below is
            # is not executed, and the test would simply pass, while
            # it shouldn't.
            #
            # So we should assert that this callback is not executed.
            # But raising an exception here leads again to the
            # execution of check_error_details() -- for this exception.
            # So we can't simply call self.fail(), but we can raise
            # a custom exception and we can check in the error
            # callback that this exception was _not_ raised.
            raise RegularCallbackExecuted

        def check_error_details(failure):
            if isinstance(failure.value, RegularCallbackExecuted):
                self.fail('Response was not an HTTP error response.')
            if not isinstance(failure, Failure):
                raise failure
            self.assertEqual('404', failure.value.status)
            self.assertEqual(
                '<html><head><title>Error handling request</title></head>\n'
                '<body><h1>Error handling request</h1>'
                'No results found: No keys found</body></html>',
                failure.value.response)

        d.addCallback(regular_execution_callback)
        return d.addErrback(check_error_details)

    def test_index_lookup(self):
        # A key index lookup form via GET.
        return self.assertContentMatches(
            '/pks/lookup?fingerprint=on&op=index&search=0xDFD20543',
            '''\
<html>
...
<title>Results for Key 0xDFD20543</title>
...
pub   1024D/DFD20543 2005-04-13
uid                  Sample Person &lt;sample.person@canonical.com&gt;
uid                  Sample Person &lt;test@canonical.com&gt;
sub   1024g/2D28D2AB 2005-04-13
sub   1024D/02BA5EF6 2005-08-01
sub   1024D/6D7DDB25 2007-10-04
...
''')

    def test_content_lookup(self):
        # A key content lookup form via GET.
        return self.assertContentMatches(
            '/pks/lookup?fingerprint=on&op=get&'
            'search=0xA419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            '''\
<html>
...
<title>Results for Key 0xA419AE861E88BC9E04B9C26FBA2B9389DFD20543</title>
...
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1.4.9 (GNU/Linux)
<BLANKLINE>
mQGiBEJdmOcRBADkNJPTBuCIefBdRAhvWyD9SSVHh8GHQWS7l9sRLEsirQkKz1yB
...
''')

    def test_lookup_key_id(self):
        # We can also request a key ID instead of a fingerprint, and it will
        # glob for the fingerprint.
        return self.assertContentMatches(
            '/pks/lookup?fingerprint=on&op=get&search=0xDFD20543',
            '''\
<html>
...
<title>Results for Key 0xDFD20543</title>
...
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1.4.9 (GNU/Linux)
<BLANKLINE>
mQGiBEJdmOcRBADkNJPTBuCIefBdRAhvWyD9SSVHh8GHQWS7l9sRLEsirQkKz1yB
...
''')

    def test_nonexistent_key(self):
        # If we request a nonexistent key, we get a nice error.
        return self.assertRaises404ErrorForKeyNotFound(
            '/pks/lookup?fingerprint=on&op=get&search=0xDFD20544')

    def test_add_key(self):
        # A key submit form via POST (see doc/gpghandler.txt for more
        # information).
        return self.assertContentMatches(
            '/pks/add',
            '''\
<html>
...
<title>Submit a key</title>
...
''')

    def postKeyText(self, keytext):
        d = self.makeService()
        path = "/pks/add"
        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "Accept": "text/plain",
        }

        def service_started(port):
            self.addCleanup(port.stopListening)
            url = 'http://localhost:%s/%s' % (
                port.getHost().port,
                path.lstrip('/'))
            params = urllib.urlencode({'keytext': keytext})
            return getPage(url, method='POST', postdata=params,
                           headers=headers)
        return d.addCallback(service_started)

    def assertResponseHasErrorCode(self, d, expected_code):
        def regular_execution_callback(content):
            # A really Twisted(tm) error check:
            #
            # This callback should _not_ be called, because setting
            # the HTTP status code to 400 counts as an error.
            # On the other hand, if the status code is
            # _not_ set, the callback check_error_details below is
            # is not executed, and the test would simply pass, while
            # it shouldn't.
            #
            # So we should assert that this callback is not executed.
            # But raising an exception here leads again to the
            # execution of check_error_details() -- for this exception.
            # So we can't simply call self.fail(), but we can raise
            # a custom exception and we can check in the error
            # callback that this exception was _not_ raised.
            raise RegularCallbackExecuted

        def check_error_details(failure):
            if isinstance(failure.value, RegularCallbackExecuted):
                self.fail('Response was not an HTTP error response.')
            if not isinstance(failure, Failure):
                raise failure
            self.assertEqual(expected_code, failure.value.status)

        d.addCallback(regular_execution_callback)
        return d.addErrback(check_error_details)

    def test_post_key_with_bad_key_text_returns_400(self):
        d = self.postKeyText('sdfsdf')
        return self.assertResponseHasErrorCode(d, '400')

    def test_post_keytext_returns_200(self):
        d = self.postKeyText('''\
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1

mQINBE/H0u0BEAC/77qha0K4BlOw67/3D3oSxN3KSa3kW6zI2Gt5zH6hHOkvIOX6
UlI2Hfku4mkg3o4urI1yL5uuMJikvG6aBhN5lWz11QG6PYHgn0EbjBw5VVhMZb17
7TOphVUypacfW6E4Imzlh7RVFFjIoVsUgOzBz0XUp3feIYLq3R4PpOgqPGgfn6zW
Fz/Htc3wEedizBB17WxjticMqry7TDn6yLGhvnCLeRU9Td94labBCA0ugwfXoAcK
+LSwPXcWpXyH/wp+SjA0ftq3hw+Oapt5tcBH+6Ym7uOiVy0LuVAwVY5+6NAnBtut
RCdsGXtxmRAvJk0s4mjf4p+R0h0LQZ5zEdMPrIJRyJZFtKQFVPDEBiRmFMk+1Ygj
JILEr7JA92lwVpLYbi/Psxp1MAUWdU6QD2vWWzPvEhCikToEXi7E5j0OMoZhni2N
wgT2LmjcGpfHo/rDR+aL+in+1Ii4Oa08Dd8y2oFF3tY9bOrigxcA0BEkoFe+d+aS
CG9BQT2ETWgssG69KGqDbGlrSgRnTxoW7yxZh9P3NK7A9YsYasNhezkTwu4QPrqm
rDTSg+a7QN69v4uXzUWONvgYsbLs0sh6Zd7upewd36uXvFbf1n4HybBvDzfsam4w
8YOmfoB1suHoDwtatnxZ0eklosmsHHhN57X3H2BekvOqNU+Doh6lHhHxLQARAQAB
tC1UaG9taSBSaWNoYXJkcyA8dGhvbWkucmljaGFyZHNAY2Fub25pY2FsLmNvbT6J
AjgEEwECACIFAk/H0u0CGwMGCwkIBwMCBhUIAgkKCwQWAgMBAh4BAheAAAoJELCM
jm0FjU8p/3oQALq/cUuxVv6Rym860G9R16se85wIhkisZ+oN6IgYfUdGS8nvujlm
3uCypNk33bfLAGIcXpyP7VjsiIhY4xzqCmexJy9dhT/E9O7MDW46qbOKUZ2SuYul
A2LW6ybDItTAyE6UUpYGG8JLjp8FkkRoIt7QWBYZb9LO6GoCCyS9aj0EIeKRvx7o
1fJWIgGrYxXb5yKfYT94JSsVCUAt6JXCuqRNupVoT9W64FKgREQ97rSDP9q0/IHA
OKt/f8nxnrpg48bkDcBpDdqdtLCdv1bzLLwaavDph+KcZqe0iJ+brJJ09+kya1sT
WFVa89+B4IcZJJbHiYEeIuLr0Da7wWPwWzPvihyTYs92at1UDbdFmo563tjdyaqf
sIYSlEIQxT0rBdXKpOeRB41UUUOjg7PQSLlNIYH7CgMLEiF8XcYApCDzTErcVnWS
ogOQqv0nCzqOAHag+5Vavaajvmn4J4nJMtva+BjdNj2Rq/mpgP+88nooFbbXTnuh
ZB/dLDh/GToHKWcg159NgUlp7l/klAawFBbn5kb1ZWU4xwn21EbVJKkcDHjbnyuH
Uzmn5wp8t1LXkg4D07E47cDVBIRISMaWfNCn1DdaKttjtOQ9G5cRJG1ohhS5N0Dg
1Snq0ckjY2ok3xowyevfOrQJMdQAIx99v9z7dyd/T1ke9YP6I/znl0KduQINBE/H
0u0BEADnZVxCTf4iZxJ7uQsA9ZTwwGSr4G/tMFxDnlpI4xb2yg42rkAjAMvfLPyX
yRPm6Nld+llqqAlhNrFepGLl0Ywg+ldLCSOOWaMHdBgqlixNLpYAjOH/47KUuXhI
hTYETt6KOlUc98Q0zN9HwH7bkT/mH6ABn8blEuDrVSMeftQ1Fklp7YrHF+Cj3ElS
ic080UkNE6CZF7EkxKyrPLfWJzASX1g8XtFTdYL6BpA4WZOSqF+tfLT735FOrGe4
B7PTvZZCl2g7/Zs1HfyRx0YTV5h7GhQnkhptrEvliGPFBoDhEj403UIqR5aODWhI
9mrTaO78d6CQIl1uUeOUSsgqItAi1X36bPyMA/t7HAmBs3Mdves6SdYa0JltsRat
qjPxGPIyHjTwu2G4S3GQQjAtaApMhE/0xP5b9IUcIuBmBir6QJuxtP2GwPrpDnj3
RJryptHfZglkemENdrzapT7Trt6KuU5y9Qfmjg6vAJn17TOOVRSd+VAhYXDTjJnR
nedgPwKM+bDvyRDvEIqzdh9pk+vGpGsqyj9Ik/0SvYmKoeQN2S9V9AGw4464/LaU
qGPerw9r7CNQU705J6wfB/CRd6cuEEOvCSjur6bNMgiEF8JzWEHeEoRUFmhUHaDT
KVjrPRT/o1w9KinxZ3+wsJi8Wxtoy8e/BR94GKGiEs2+/AdkPQARAQABiQIfBBgB
AgAJBQJPx9LtAhsMAAoJELCMjm0FjU8pXgsQALU2hzSRbd2Y6UIVg0LsX9xiRYxT
1lis0Md34FM+1XiGcXRd2hqacyXTBdeSRoe3vJccaGbBSpxt7KZKV1C3AC1FT5mE
HkFM1TfNyw8fD6sdXJ7kaJ9L/Y1GxRz/CHmJfdu7ZP3A8cH/E99+zk+ER9CaNAnr
g32NB+d4RORDFSRode8aDUsouqHjUy4bxoHh4Myq1Ehx2n7+jfu85pcdovVQXW7J
Tyw8+FKSIKYr4VUW1/skU6MW1j0+0/uyYbJiF98vwdzEmhz4p+IAlpOL8wwaP7kp
RU1u7jMcxha7IY9cXQJHU1D9/LeZ/WepLEJHnfTWEPGgTFBsXgXIKUIPVBREy6VR
CjgRvfpoK2q+81GCXGkYxAfa9WzSjfXm+3+S44Jpy6h+F6EjC3VWW3zoy/usHf6R
AyLUv8S1avIZ5mEcD41pDu829P5IuqdG7mJQ96MoJhtcbKs6rnSlGE7tDRf+PTq3
YMVftR3C+WhtYD8Le5/SlniNTSvvtsVtdM2EXibNVi9BRR9TqMWvD2kE4hPF3MdR
UJW5d74sGgq40YDJZqidkPXwbS0/XwGaSmMYwuNOMrOqPa/Xlp/56Rre6MJRMJDt
Gg0xyMMkCTt2V4Ah9XVEZ/YI7ZRq8qFB7hCHXhTjwT4MImlBrEtQTyNXfJNB5f3R
DC7/1JZmKIDPzXV1
=MSHM
-----END PGP PUBLIC KEY BLOCK-----
        ''')
        return d.addCallback(
            self.assertThat, Contains('Key added'))


def test_suite():
    from unittest import TestLoader
    return TestLoader().loadTestsFromName(__name__)
